@@ -375,7 +375,9 @@ typedef uint mrn_alter_table_flags;
 #  define MRN_HANDLER_HAVE_CAN_SWITCH_ENGINES
 #endif
 
-#if !defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80016
+#if !defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80027
+/* nothing */
+#elif !defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 80016
 #  define MRN_HANDLER_COND_PUSH_HAVE_OTHER_TABLES_OK
 #else
 #  define MRN_HANDLER_HAVE_COND_POP
@@ -868,9 +870,6 @@ protected:
 #  endif
 
 private:
-  void mkdir_p(const char *directory);
-  ulonglong file_size(const char *path);
-
   bool have_unique_index();
 
   bool is_foreign_key_field(const char *table_name,
